/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "framework.h"

namespace mysdl
{
	/**
	 * The current application. Automatically set by Application's
	 * constructor.
	 */
	Application *application;

	// ================================================================
	//  ApplicationState
	// ================================================================

	/*
	 * See header file for documentation.
	 */
	ApplicationState::ApplicationState() :
		terminate(false),performStageCount(0),drawStageCount(0)
	{
	}

	/*
	 * See header file for documentation.
	 */
	ApplicationState::~ApplicationState()
	{
		deactivate();
	}

	/*
	 * See header file for documentation.
	 */
	void ApplicationState::deactivate()
	{
		if(application->state==this)
		{
			application->state=0;
		}
	}

	/*
	 * See header file for documentation.
	 */
	void ApplicationState::activate()
	{
		if(application->state)
		{
			application->state->deactivate();
		}

		application->state=this;
	}

	/*
	 * See header file for documentation.
	 */
	void ApplicationState::perform(Uint32 ticks)
	{
		for(int i=0; (!terminate)&&(i<performStageCount); i++)
		{
			doPerform(i,ticks);
		}

		if(terminate)
		{
			delete this;
		}
	}

	/*
	 * See header file for documentation.
	 */
	void ApplicationState::draw()
	{
		for(int i=0; i<drawStageCount; i++)
		{
			doDraw(i);
		}
	}

	// ================================================================
	//  Application
	// ================================================================

	/*
	 * See header file for documentation.
	 */
	Application::Application() :
		state(0),queue(0)
	{
		application=this;
	}

	/*
	 * See header file for documentation.
	 */
	Application::~Application()
	{
		mysdl::quit();
	}

	/*
	 * See header file for documentation.
	 */
	void Application::run()
	{
		getElapsedTicks();

		while(state)
		{
			perform(getElapsedTicks());
		}
	}

	/*
	 * See header file for documentation.
	 */
	void Application::handleEvents(ControlQueue *queue)
	{
		if(!(queue->readEvents()))
		{
			delete state;
		}

		while(state)
		{
			Trigger *e;

			(*queue) >> e;

			if(e)
			{
				state->handleEvent(e);
			}
			else
			{
				break;
			}
		}
	}

	/*
	 * See header file for documentation.
	 */
	void Application::perform(Uint32 ticks)
	{
		if((state)&&(queue))
		{
			handleEvents(queue);
		}

		if(state)
		{
			state->perform(ticks);
		}

		if(state)
		{
			state->draw();
		}
	}

	// ================================================================
	//  FrameRateLimitedApplication
	// ================================================================

	FrameRateLimitedApplication::FrameRateLimitedApplication() :
			Application(),
	maxTicks(0),desiredTicks(0)
	{
	}

	/*
	* See header file for documentation.
	*/
	void FrameRateLimitedApplication::perform(Uint32 ticks)
	{
		if((maxTicks>0)&&(ticks>maxTicks))
		{
			ticks=maxTicks;
			SDL_Delay(1); // ?
		}
		else if((desiredTicks>0)&&(ticks<desiredTicks))
		{
			SDL_Delay(desiredTicks-ticks);
			ticks+=getElapsedTicks();
		}

		Application::perform(ticks);
	}

	// ================================================================
	//  FixedRateApplication
	// ================================================================

	FixedRateApplication::FixedRateApplication() :
		Application(),
		tickBuffer(0),fixedTicks(100)
	{
	}

	/*
	* See header file for documentation.
	*/
	void FixedRateApplication::perform(Uint32 ticks)
	{
		tickBuffer+=ticks;

		while(tickBuffer>=fixedTicks)
		{
			tickBuffer-=fixedTicks;

			if((state)&&(queue))
			{
				handleEvents(queue);
			}

			if(state)
			{
				state->perform(fixedTicks);
			}
		}

		if(state)
		{
			state->draw();
		}
	}
};
