/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "animations.h"

namespace mysdl
{
	namespace gfx
	{
		// ================================================================
		//  AnimationResourceBankGenerator
		// ================================================================

		/**
		 * A resource bank generator for animation banks.
		 */
		class AnimationResourceBankGenerator : public ResourceBankGenerator
		{
			public:
				BasicResourceBank *generateBank(const std::string &name) const;
		};

		/**
		 * Generates an animation bank if the bank name is "animations".
		 *
		 * @param name Name for the bank that should be generated.
		 *
		 * @return An animation bank, or 0.
		 */
		BasicResourceBank *AnimationResourceBankGenerator::generateBank(const std::string &name) const
		{
			if(name=="animations")
			{
				return(new AnimationBank());
			}
			else
			{
				return(0);
			}
		}

		/** A generator for animation banks (automatically created &
		 *  registered. */
		AnimationResourceBankGenerator animationResourceBankGenerator;

		// ================================================================
		//  XML constants
		// ================================================================

		/** XML element and attribute names. */
		const char *const XML_ANIMATIONS_ANIMATIONS="animations";
		const char *const XML_ANIMATIONS_ANIMATION ="animation";
		const char *const XML_ANIMATIONS_NAME      ="name";
		const char *const XML_ANIMATIONS_FRAME     ="frame";
		const char *const XML_ANIMATIONS_IMAGE     ="image";
		const char *const XML_ANIMATIONS_DELAY     ="delay";
		const char *const XML_ANIMATIONS_BACK      ="back";
		const char *const XML_ANIMATIONS_IMAGEBANK ="imageBank";

		const char *const XML_ANIMATIONS_LOOP      ="loop";
		const char *const XML_ANIMATIONS_VAR       ="var";
		const char *const XML_ANIMATIONS_FROM      ="from";
		const char *const XML_ANIMATIONS_TO        ="to";
		const char *const XML_ANIMATIONS_DIGITS    ="digits";
		const char *const XML_ANIMATIONS_STEP      ="step";

		// ================================================================
		//  AnimationFrame
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		AnimationFrame::AnimationFrame() : image(),delay()
		{
		}

		/*
		 * See header file for documentation.
		 */
		AnimationFrame::AnimationFrame(Image *image,GLfloat delay) : image(image),delay(delay)
		{
		}

		// ================================================================
		//  Animation
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Animation::Animation() : NamedResource<AnimationBank>(),delay(10)
		{
		}

		/*
		 * See header file for documentation.
		 */
		Animation::Animation(const std::string &name) : NamedResource<AnimationBank>(name),delay(10)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void Animation::getFromNode(const xml::Node *root,ImageBank *images,std::list<mysdl::xml::LoopData *> loops)
		{
			//debug::addMessage("Animation::getFromNode()");

			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				const xml::Node *node=*itr;

				if(node->name==XML_ANIMATIONS_LOOP)
				{
					const std::string *var=node->getAttribute(XML_ANIMATIONS_VAR);

					if(!var)
					{
						debug::error(dbgAnimations,"No variable specified for loop. Skipping loop.");
						continue;
					}

					int min   =(int)(node->getAttributeLong(XML_ANIMATIONS_FROM,0));
					int max   =(int)(node->getAttributeLong(XML_ANIMATIONS_TO,0));
					int step  =(int)(node->getAttributeLong(XML_ANIMATIONS_STEP,1));
					int digits=(int)(node->getAttributeLong(XML_ANIMATIONS_DIGITS,0));

					if(step==0)
					{
						debug::error(dbgAnimations,"Step size 0 invalid: %s. Skipping loop.",var->c_str());
					}

					if(((step>0)&&(max<=min))||((step<0)&&(max>=min)))
					{
						debug::error(dbgAnimations,"Loop range invalid: %s [%d;%d] (step: %d). Skipping loop.",var->c_str(),min,max,step);
						continue;
					}

					mysdl::xml::LoopData *entry=new mysdl::xml::LoopData(*var,min,max,step,digits);
					loops.push_back(entry);

					do
					{
						getFromNode(node,images,loops);
					}
					while(entry->advance());

					loops.pop_back();
					delete entry;
				}
				else if(node->name==XML_ANIMATIONS_FRAME)
				{
					const std::string *imageNamePtr=node->getAttribute(XML_ANIMATIONS_IMAGE);

					if(!imageNamePtr)
					{
						debug::error(dbgImages,"No name specified for animation frame image. Skipping.");
						continue;
					}

					std::string imageName=*imageNamePtr;

					for(std::list<mysdl::xml::LoopData *>::iterator itr=loops.begin(); itr!=loops.end(); itr++)
					{
						std::string varStr("["+toStr((*itr)->var)+"]");
						std::string valStr((*itr)->getString());

						for(std::string::size_type i=imageName.find(varStr); i!=std::string::npos; i=imageName.find(varStr))
						{
							imageName.replace(i,varStr.size(),valStr);
						}
					}

					GLfloat delay=node->getAttributeFloat(XML_ANIMATIONS_DELAY,this->delay);

					frames.push_back(AnimationFrame((*images)[imageName],delay));
				}
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  AnimationReference
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		AnimationReference *Animation::getReference(GLfloat delay)
		{
			return(new AnimationReference(this,delay));
		}

		/*
		 * See header file for documentation.
		 */
		AnimationReference::AnimationReference() :
			Image(),frame(0),animation(0),speed(1),index(0),timeIndex(0),target(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		AnimationReference::AnimationReference(Animation *animation,GLfloat speed) :
			Image(),frame(0),animation(animation),speed(speed),index(0),timeIndex(0),target(0)
		{
			//debug::addMessage("AnimationReference::AnimationReference()");

			index=0;

			if(animation)
			{
				target=animation->frames.size()-(animation->loop ? 0 : 1);
			}
			else
			{
				target=0;
			}

			setFrame(animation ? &(animation->frames[0]) : 0);

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationReference::setFrame(AnimationFrame *frame)
		{
			//debug::addMessage("AnimationReference::setFrame()");

			this->frame=frame;

			if(frame)
			{
				image=frame->image;
				width=image->getWidth();
				height=image->getHeight();
				bitmask=image->getBitmask();
				hotspot=image->getHotspot();
				outerBoundingBox=*(image->getOuterBoundingBox());
				innerBoundingBox=*(image->getInnerBoundingBox());
			}
			else
			{
				image=0;
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationReference::draw(const Vector2D &pos,BlendMode blendMode) const
		{
			//debug::addMessage("AnimationReference::draw(const Vector2D&)");
			image->draw(pos,blendMode);
			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationReference::draw(const Rect &pos,BlendMode blendMode) const
		{
			//debug::addMessage("AnimationReference::draw(const Rect&)");
			image->draw(pos,blendMode);
			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationReference::drawStretched(const Rect &pos,BlendMode blendMode) const
		{
			//debug::addMessage("AnimationReference::drawStretched(const Rect&)");
			image->draw(pos,blendMode);
			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationReference::perform(GLfloat ticks)
		{
			//debug::addMessage("AnimationReference::perform()");

			if(index!=target)
			{
				timeIndex+=ticks;

				while(timeIndex>speed*frame->delay)
				{
					if(index!=target)
					{
						index+=(target<index ? -1 : +1);

						if(index<0)
						{
							index=(animation->frames.size()-1);
						}
						else if(index>=animation->frames.size())
						{
							index=0;
						}

						setFrame(&(animation->frames[index]));
					}

					timeIndex-=speed*frame->delay;
				}
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		bool AnimationReference::isFinished()
		{
			return((index==target)&&(!(animation->loop)));
		}

		// ================================================================
		//  AnimationBank
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		AnimationBank::~AnimationBank()
		{
			mysdl::debug::debug(dbgAnimations,"Deleting animation bank %p.",this);
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationBank::getFromNode(const xml::Node *root,const std::string &baseDir)
		{
			getFromNode(root,0);
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationBank::getFromNode(const xml::Node *root,ImageBank *images)
		{
			//debug::addMessage("AnimationBank::getFromNode()");

			if(manager)
			{
				if(const std::string *bankName=root->getAttribute(XML_ANIMATIONS_IMAGEBANK))
				{
					ImageBank *bank=(ImageBank *)((*manager)[*bankName]);

					if(bank)
					{
						images=bank;
					}
				}
			}

			getContentsFromNode(root,images);

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void AnimationBank::getContentsFromNode(const xml::Node *root,ImageBank *images)
		{
			static std::list<mysdl::xml::LoopData *> loops; // \note: Not thread safe!

			//debug::addMessage("AnimationBank::getContentsFromNode()");

			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				const xml::Node *node=*itr;

				if(node->name==XML_ANIMATIONS_LOOP)
				{
					const std::string *var=node->getAttribute(XML_ANIMATIONS_VAR);

					if(!var)
					{
						debug::error(dbgAnimations,"No variable specified for loop. Skipping loop.");
						continue;
					}

					int min 	=(int)(node->getAttributeLong(XML_ANIMATIONS_FROM,0));
					int max 	=(int)(node->getAttributeLong(XML_ANIMATIONS_TO,0));
					int step	=(int)(node->getAttributeLong(XML_ANIMATIONS_STEP,1));
					int digits=(int)(node->getAttributeLong(XML_ANIMATIONS_DIGITS,0));

					if(step==0)
					{
						debug::error(dbgAnimations,"Step size 0 invalid: %s. Skipping loop.",var->c_str());
					}

					if(((step>0)&&(max<=min))||((step<0)&&(max>=min)))
					{
						debug::error(dbgAnimations,"Loop range invalid: %s [%d;%d] (step: %d). Skipping loop.",var->c_str(),min,max,step);
						continue;
					}

					mysdl::xml::LoopData *entry=new mysdl::xml::LoopData(*var,min,max,step,digits);
					loops.push_back(entry);

					do
					{
						getFromNode(node,images);
					}
					while(entry->advance());

					loops.pop_back();
					delete entry;
				}
				else if(node->name==XML_ANIMATIONS_ANIMATION)
				{
					const std::string *namePtr=node->getAttribute(XML_ANIMATIONS_NAME);

					if(!namePtr)
					{
						debug::error(dbgImages,"No name specified for animation. Skipping.");
						continue;
					}

					std::string name=*namePtr;

					for(std::list<mysdl::xml::LoopData *>::iterator itr=loops.begin(); itr!=loops.end(); itr++)
					{
						std::string varStr("["+toStr((*itr)->var)+"]");
						std::string valStr(toStr((*itr)->value));

						for(std::string::size_type i=name.find(varStr); i!=std::string::npos; i=name.find(varStr))
						{
							name.replace(i,varStr.size(),valStr);
						}
					}

					ImageBank *currentBank=images;

					if(manager)
					{
						if(const std::string *bankName=root->getAttribute(XML_ANIMATIONS_IMAGEBANK))
						{
							ImageBank *bank=(ImageBank *)((*manager)[*bankName]);

							if(bank)
							{
								currentBank=bank;
							}
						}
					}

					Animation *animation=new Animation(name);

					animation->delay=node->getAttributeFloat(XML_ANIMATIONS_DELAY,animation->delay);
					animation->loop=node->getAttributeBool(XML_ANIMATIONS_LOOP,animation->loop);

					try
					{
						animation->getFromNode(node,currentBank,loops);
						(*this)+=animation;
					}
					catch(std::string ex)
					{
						debug::error(dbgAnimations,"Error reading animation \"%s\": %s",animation->name.c_str(),ex.c_str());
						delete animation;
					}
				}
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		std::string AnimationBank::getNodeName()
		{
			return(XML_ANIMATIONS_ANIMATIONS);
		}
	};
};
