/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <signal.h>
#include <stdlib.h>
#include <iostream>

#include "SDL/SDL_image.h"
#include "SDL/SDL_opengl.h"
//#include "glext.h"

#include "png.h"

#include "buildinfo.h"
#include "debug.h"
#include "mysdl.h"
#include "resources.h"

namespace mysdl
{
	/** XML tag and attribute names for colours. */
	static const std::string XML_COLOR_CODE ="code";
	static const std::string XML_COLOR_RED  ="red";
	static const std::string XML_COLOR_GREEN="green";
	static const std::string XML_COLOR_BLUE ="blue";
	static const std::string XML_COLOR_ALPHA="alpha";

	/** String representations of the JustificationHorizontal values. */
	const char *STR_JUSTIFICATION_HORIZONTAL[]={"left","center","right"};

	/** String representations of the JustificationVertical values. */
	const char *STR_JUSTIFICATION_VERTICAL[]={"top","center","bottom"};

	/** A pair of debug group identifiers and debug group names. */
	typedef std::pair<int,std::string> DebugPair;

	/** Default debug group identifiers and names. */
	static const DebugPair defDebugGroups[]=
	{
		DebugPair(dbgCore,std::string("core")),
		DebugPair(dbgControls,std::string("controls")),
		DebugPair(dbgGui,std::string("gui")),
		DebugPair(dbgDdd,std::string("ddd")),
		DebugPair(dbgMath,std::string("math")),
		DebugPair(dbgSprites,std::string("sprites")),
		DebugPair(dbgImages,std::string("images")),
		DebugPair(dbgFonts,std::string("fonts")),
		DebugPair(dbgXml,std::string("xml")),
		DebugPair(dbgAudio,std::string("audio")),
		DebugPair(dbgCollision,std::string("collision")),
		DebugPair(dbgAnimations,std::string("animations")),
		DebugPair(dbgResources,std::string("resources")),
		DebugPair(0,std::string(""))
	};

	/** Viewport representing the full screen (or application window). */
	Viewport vpScreen=Viewport();

	/** Pointer to the current viewport. */
	Viewport *currentViewport=0;

	/** The currently active render context. Should be considered read-only. */
	RenderContext *currentRenderContext;

	extern std::list<BackgroundTask *> backgroundTasks;
	std::list<BackgroundTask *> backgroundTasks;

	/** Set to <code>true</code> by the quit() function. */
	bool quitting=false;

	/** Screen surface, typically pointing to the current backbuffer. */
	SDL_Surface *screen=0;

	/** <code>true</code> if the program is running in fullscreen mode. */
	bool fullscreen=false;

	/**
	 * Converts a string into a JustificationHorizontal value.
	 *
	 * @param str String representing a horizontal justification value.
	 *
	 * @return Justification value.
	 */
	JustificationHorizontal justificationHorizontal(const char *str)
	{
		if(strcmp(str,STR_JUSTIFICATION_HORIZONTAL[jhLeft])==0)
		{
			return(jhLeft);
		}
		else if(strcmp(str,STR_JUSTIFICATION_HORIZONTAL[jhCenter])==0)
		{
			return(jhCenter);
		}
		else if(strcmp(str,STR_JUSTIFICATION_HORIZONTAL[jhRight])==0)
		{
			return(jhRight);
		}

		//return(jhLeft); // \todo Throw exception?
		throw std::string(str)+std::string(" is not a valid JustificationHorizontal value.");
	}

	/**
	 * Converts a string into a JustificationVertical value.
	 *
	 * @param str String representing a vertical justification value.
	 *
	 * @return Justification value.
	 */
	JustificationVertical justificationVertical(const char *str)
	{
		if(strcmp(str,STR_JUSTIFICATION_VERTICAL[jvTop])==0)
		{
			return(jvTop);
		}
		else if(strcmp(str,STR_JUSTIFICATION_VERTICAL[jvCenter])==0)
		{
			return(jvCenter);
		}
		else if(strcmp(str,STR_JUSTIFICATION_VERTICAL[jvBottom])==0)
		{
			return(jvBottom);
		}

		//return(jvBottom); // \todo Throw exception?
		throw std::string(str)+std::string(" is not a valid JustificationVertical value.");
	}

	/**
	 * Creates a new colour, reading its data from the specified
	 * DOM node.
	 *
	 * @param node DOM node containing colour data.
	 */
	Color::Color(const xml::Node *node) :
		r(0),g(0),b(0),a(0)
	{
		const std::string *s=node->getAttribute(XML_COLOR_CODE);

		if(s)
		{
			set(s->c_str());
		}
		else
		{
			r=node->getAttributeFloat(XML_COLOR_RED,0);
			g=node->getAttributeFloat(XML_COLOR_GREEN,0);
			b=node->getAttributeFloat(XML_COLOR_BLUE,0);
			a=node->getAttributeFloat(XML_COLOR_ALPHA,1);
		}
	}

	/**
	 * Sets the colour as the current colour.
	 */
	void Color::apply() const
	{
		glColor4fv(&r);
	}

	/**
	 * Initializes MySDL and the specified subsystems.
	 *
	 * @param systems Sub systems which shall be initialized.
	 *
	 * @param debug   Whether to enable MySDL debugging features, like
	 *                setting the default signal handler (recommended).
	 *
	 * @throws std::string When an error occurs.
	 */
	void init(SubSystem systems,bool debug)
	{
		currentViewport=0;

		for(int i=0; defDebugGroups[i].first!=0; i++)
		{
			debug::groups[defDebugGroups[i].first].name=std::string("mysdl:")+defDebugGroups[i].second;
		}

		debug::init();

		if(debug)
		{
			signal(SIGSEGV,handleSignal);
			signal(SIGILL ,handleSignal);
			signal(SIGFPE ,handleSignal);
			signal(SIGINT ,handleSignal);
			signal(SIGTERM,handleSignal);
			signal(SIGABRT,handleSignal);
		}

		debug::info(dbgCore,"MySDL version %s.",getFullVersion().c_str());

		fullscreen=false;

		if(systems>ssNone)
		{
			if(SDL_Init(systems|SDL_INIT_NOPARACHUTE)!=0)
			{
				throw std::string(SDL_GetError());
			}

			SDL_EnableUNICODE(1); // \todo only activate when character input is required -> GUI
		}
	}

	/**
	 * Frees all MySDL resources.
	 */
	void cleanUp()
	{
		showCursor(true);

		while(!(backgroundTasks.empty()))
		{
			delete backgroundTasks.back();
			backgroundTasks.pop_back();
		}
	}

	/**
	 * Cleanly terminates the program, freeing all MySDL resources.
	 */
	void quit()
	{
		if(!quitting)
		{
			quitting=true;
			cleanUp();
		}
		else
		{
			debug::warning(dbgCore,"Quit called again.");
		}

		SDL_Quit();
		exit(0);
	}

	void setResolution(int width,int height,int bpp,bool fullscreen,int depth,int stencil)
	{
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE,bpp/4);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,bpp/4);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,bpp/4);
		SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,bpp/4);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);

		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,depth);
		SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,stencil);

		if(!(screen=SDL_SetVideoMode(width,height,bpp,SDL_OPENGL|(fullscreen ? SDL_FULLSCREEN : 0))))
		{
			throw std::string(SDL_GetError());
		}

		debug::info(dbgCore,"MySDL is now using OpenGL version \"%s\" at %dx%dx%d.",glGetString(GL_VERSION),width,height,bpp);

		vpScreen.pos.x=0;
		vpScreen.pos.y=0;
		vpScreen.dimensions.x=width;
		vpScreen.dimensions.y=height;

		vpScreen.apply();

		mysdl::fullscreen=fullscreen;

		glClearColor(0,0,0,0);
	}

	/**
	 * Moves the drawing cursor to the origin.
	 */
	void resetCursor()
	{
		glLoadIdentity();
	}

	/**
	 * Moves the drawing cursor by the specified 2D vector.
	 *
	 * @param v Vector.
	 */
	void moveCursor(const Vector2D &v)
	{
		glTranslatef(v.x,v.y,0);
	}

	/**
	 * Moves the drawing cursor by the specified 3D vector.
	 *
	 * @param v Vector.
	 */
	void moveCursor(const Vector3D &v)
	{
		glTranslatef(v.x,v.y,v.z);
	}

	/**
	 * Rotates the drawing cursor.
	 *
	 * @param angle Rotation angle in degrees.
	 * @param v     Rotation axis.
	 */
	void rotateCursor(GLfloat angle,const Vector3D &v)
	{
		glRotatef(angle,v.x,v.y,v.z);
	}

	/**
	 * Creates a new render context, using default values.
	 */
	RenderContext::RenderContext() :
		wireframe(false),shadeModel(smFlat),
		qualityFog(0),qualityPerspective(0),qualityPoints(0),qualityPolygons(0),
		lighting(0),ambient(0.2,0.2,0.2,0.1)
	{
	}

	/**
	 * Activates the render context.
	 */
	void RenderContext::apply()
	{
		currentRenderContext=this;

		glPolygonMode(GL_FRONT_AND_BACK,wireframe ? GL_LINE : GL_FILL);

		glShadeModel(shadeModel==smFlat ? GL_FLAT : GL_SMOOTH);

		glHint(GL_FOG_HINT,qualityFog ? (qualityFog>0 ? GL_NICEST : GL_FASTEST) : GL_DONT_CARE);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,qualityPerspective ? (qualityPerspective>0 ? GL_NICEST : GL_FASTEST) : GL_DONT_CARE);
		glHint(GL_POINT_SMOOTH_HINT,qualityPoints ? (qualityPoints>0 ? GL_NICEST : GL_FASTEST) : GL_DONT_CARE);
		glHint(GL_POLYGON_SMOOTH_HINT,qualityPolygons ? (qualityPolygons>0 ? GL_NICEST : GL_FASTEST) : GL_DONT_CARE);

		if(lighting)
		{
			glEnable(GL_LIGHTING);
			glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
			glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL,qualityLight>0 ? GL_SEPARATE_SPECULAR_COLOR : GL_SINGLE_COLOR);
			glLightModelfv(GL_LIGHT_MODEL_AMBIENT,&(ambient.r));
		}
		else
		{
			glDisable(GL_LIGHTING);
		}
	}

	/**
	 * Creates a new 2D render context, using default values.
	 */
	RenderContext2D::RenderContext2D() :
		RenderContext(),
		virtualDimensions(),blending(false),blendMode(bmDefault)
	{
		if(currentViewport)
		{
			virtualDimensions=currentViewport->dimensions;
		}
	}

	/**
	 * Creates a new 2D render context, using the specified virtual
	 * dimensions. 0/0 will be mapped to the lower left corner of
	 * the viewport, while the specified vector corresponds to the
	 * upper right corner.
	 * \todo check: really corresponding to *viewport*?
	 *
	 * @param virtualDimensions Virtual dimensions for this viewport.
	 */
	RenderContext2D::RenderContext2D(Vector2D virtualDimensions) :
		RenderContext(),virtualDimensions(virtualDimensions),blending(false),blendMode(bmDefault)
	{
	}

	/**
	 * Activates the render context. This sets the unit dimensions
	 * according to the current viewport.
	 */
	void RenderContext2D::apply()
	{
		RenderContext::apply();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0,virtualDimensions.x,0,virtualDimensions.y);
		glMatrixMode(GL_MODELVIEW);

		glDisable(GL_DEPTH_TEST);

		if(blending)
		{
			glEnable(GL_BLEND);
			//glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); // \todo make customisable
			setBlendMode(blendMode);
		}
		else
		{
			glDisable(GL_BLEND);
		}

		glDisable(GL_NORMALIZE);
		glDisable(GL_CULL_FACE);
		glDisable(GL_FOG);
	}

	/**
	 * Creates a new 3D render context, using default values.
	 */
	RenderContext3D::RenderContext3D() :
		RenderContext(),
		depthTest(true),clearDepth(10000.0),
		fogModel(fmNone),fogStart(0),fogEnd(0),fogDensity(0),qualityFog(0),fogColor()
	{
		//\todo glDepthFunc?
	}

	/**
	 * Activates the render context.
	 */
	void RenderContext3D::apply()
	{
		RenderContext::apply();

		glClearDepth(clearDepth);

		if(depthTest)
		{
			glEnable(GL_DEPTH_TEST);
		}
		else
		{
			glDisable(GL_DEPTH_TEST);
		}

		glDisable(GL_BLEND); // \todo better

		glEnable(GL_NORMALIZE);

		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);

		if(fogModel==fmNone)
		{
			glDisable(GL_FOG);
		}
		else
		{
			glFogi(GL_FOG_MODE,(fogModel==fmLinear ? GL_LINEAR : fogModel==fmExponential ? GL_EXP : GL_EXP2));
			glFogf(GL_FOG_START,fogStart);
			glFogf(GL_FOG_END,fogEnd);
			glFogf(GL_FOG_DENSITY,fogDensity);
			glFogfv(GL_FOG_COLOR,&(fogColor.r));
			glHint(GL_FOG_HINT,qualityFog ? (qualityFog>0 ? GL_NICEST : GL_FASTEST) : GL_DONT_CARE);
			glEnable(GL_FOG);
		}
	}

	/**
	 * Sets the program window's caption.
	 *
	 * @param caption New window caption.
	 */
	void setWindowCaption(char *const caption)
	{
		SDL_WM_SetCaption(caption,caption);
	}

	/**
	 * Sets the program window's icon.
	 * 
	 * @param imageFileName Name of the image file to be used as the icon.
	 */
	void setWindowIcon(const char *imageFileName)
	{
		SDL_WM_SetIcon(IMG_Load(imageFileName),0);
	}
	
	/**
	 * Shows or hides the cursor within the program's window.
	 *
	 * @param show <code>true</code> to show,
	 *             <code>false</code> to hide the cursor.
	 */
	void showCursor(bool show)
	{
		SDL_ShowCursor(show ? SDL_ENABLE : SDL_DISABLE);
	}

	/**
	 * Enables or disables double buffering.
	 *
	 * @param enable <code>true</code> to enable,
	 *               <code>false</code> to disable double buffering.
	 */
	void enableDoubleBuffer(bool enable)
	{
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,enable ? 1 : 0);
	}

	/**
	 * Specifies the colour used for clearing the screen.
	 *
	 * @param color New clear color.
	 */
	void setClearColor(const Color &color)
	{
		glClearColor(color.r,color.g,color.b,color.a);
	}

	/**
	 * Clears the screen using the colour set by setClearColor().
	 */
	void clearScreen()
	{
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT); // \todo Make customizable, including stencil
	}

	/**
	 * Enables/disables fullscreen mode. Does not work in Windows.
	 *
	 * @param fullscreen <code>true</code> To switch to fullscreen mode,
	 *                   <code>false</code> for windowed mode.
	 */
	void setFullscreen(bool fullscreen)
	{
		if(mysdl::fullscreen!=fullscreen)
		{
			toggleFullscreen();
		}
	}

	/**
	 * Enables/disableds fullscreen mode. Does not work in Windows.
	 * \todo *crashes* in Windows??
	 */
	void toggleFullscreen()
	{
		if(SDL_WM_ToggleFullScreen(screen))
		{
			fullscreen=!fullscreen;
		}
		else
		{
			throw std::string("Could not toggle fullscreen mode: ")+std::string(SDL_GetError());
		}
	}

	/**
	 * Swaps front and back buffers in double buffered mode.
	 */
	void flipScreen()
	{
		SDL_GL_SwapBuffers();
	}

	/**
	 * Initializes a joystick.
	 * \todo make class?
	 *
	 * @param id Joystick ID (typically 0 for the first joystick, 1 for the second etc.).
	 *
	 * @return Joystick object for the opened joystick, or <code>0</code> on error.
	 */
	Joystick *initJoystick(int id)
	{
		return(SDL_JoystickOpen(id));
	}

	/**
	 * Closes a joystick.
	 *
	 * @param joy Open joystick.
	 */
	void cleanUpJoystick(Joystick *joy)
	{
		SDL_JoystickClose(joy);
	}

	/**
	 * Saves a screenshots as a PNG file.
	 *
	 * Based on code by Ray Kelm:
	 * http://www.libsdl.org/pipermail/sdl/2000-September/030012.html
	 *
	 * ("Oh. um. This code is released for anyone to anything they want
   *  with blah blah blah and don't sue me if it destroys your life etc.")
	 *
	 * @param filename Bitmap file name.
	 *
	 * @throws std::string If an error occurs.
	 */
	void saveScreenshot(const char *filename)
	{
		SDL_Surface *temp=SDL_CreateRGBSurface(SDL_SWSURFACE,screen->w,screen->h,24,
			#if SDL_BYTEORDER==SDL_LIL_ENDIAN
				0x000000FF,0x0000FF00,0x00FF0000,0
			#else
				0x00FF0000,0x0000FF00,0x000000FF,0
			#endif
		);

		if(!temp)
		{
			throw std::string(SDL_GetError());
		}

		unsigned char *pixels=new unsigned char[3*screen->w*screen->h];

		if(!pixels)
		{
			SDL_FreeSurface(temp);
			throw std::string("Out of memory");
		}

		glReadPixels(0,0,screen->w,screen->h,GL_RGB,GL_UNSIGNED_BYTE,pixels);

		for(int i=0; i<screen->h; i++)
		{
			memcpy(((char *)temp->pixels)+temp->pitch*i,pixels+3*screen->w*(screen->h-i-1),screen->w*3);
		}

		delete[] pixels;

		savePng(temp,filename);
		SDL_FreeSurface(temp);
	}

	/**
	 * MySDL's default signal handler. Prints the signal type and
	 * current debug stack trace to stderr and then terminates
	 * the application.
	 *
	 * @param signal Signal type.
	 */
	void handleSignal(int sig)
	{
		static int count=0;

		fprintf(stderr,"MySDL parachute deployed: ");

		if((count++)>0)
		{
			signal(SIGSEGV,0);
			signal(SIGILL ,0);
			signal(SIGFPE ,0);
			signal(SIGINT ,0);
			signal(SIGTERM,0);
			signal(SIGABRT,0);

			fprintf(stderr,"Recursive call.\n");
		}
		else
		{
			switch(sig)
			{
				case SIGFPE:  fprintf(stderr,"Floating point exception.\n");
											break;
				case SIGILL:  fprintf(stderr,"Illegal instruction.\n");
											break;
				case SIGSEGV: fprintf(stderr,"Segmentation fault.\n");
											break;
				case SIGINT:  fprintf(stderr,"Interactive attention.\n");
											break;
				case SIGTERM: fprintf(stderr,"Termination request.\n");
											break;
				case SIGABRT: fprintf(stderr,"Abnormal termination (abort).\n");
											break;
				default:      fprintf(stderr,"Signal #%d.\n",sig);
											break;
			}
		}

		fflush(stderr);
		quit();
	}

	/**
	 * Gets a pixel from a surface. If the surface must be locked before its
	 * memory can be accessed, the user must lock it before calling this function
	 * and unlock it afterwards.
	 *
	 * @param surface The surface.
	 * @param x       X coordinate of requested pixel.
	 * @param y       Y coordinate of requested pixel.
	 *
	 * @return Pixel colour. The format depends on the surface format (\todo verify).
	 */
	Uint32 getPixel(SDL_Surface *surface,int x,int y)
	{
		Uint32 pixel=0;
		int bpp=surface->format->BytesPerPixel;
		Uint8 *p=(Uint8*)(((Uint8*)surface->pixels)+(y*surface->pitch)+(x*bpp));

		if(bpp==1)
		{
			pixel=*p;
		}
		else if(bpp==2)
		{
			pixel=*(Uint16*)p;
		}
		else if(bpp==3)
		{
			if(SDL_BYTEORDER==SDL_BIG_ENDIAN)
			{
				pixel=p[0] << 16 | p[1] << 8 | p[2];
			}
			else
			{
				pixel=p[0] | p[1] << 8 | p[2] << 16;
			}
		}
		else if(bpp == 4)
		{
			if(SDL_BYTEORDER==SDL_BIG_ENDIAN)
			{
				pixel=p[0] << 24 | p[1] << 16 | p[2] << 8 | p[3];
			}
			else
			{
				pixel=p[0] | p[1] << 8 | p[2] << 16 | p[3] << 24;
			}
		}
		else
		{
			mysdl::debug::critical(mysdl::dbgCore,"Unknown surface format (bpp=%d).",bpp);
		}

		return(pixel);
	}

	/**
	 * Saves an SDL surface to an RGBA PNG file. Intended for internal
	 * use and helper applications, since it requires explicit dependency
	 * on the SDL library.
	 *
	 * @param surf     Surface that will be saved.
	 * @param filename Image filename.
	 *
	 * @throws std::string on error.
	 */
	void savePng(SDL_Surface *surf,const char *filename)
	{
		png_structp png_ptr;
		png_infop info_ptr;
		Uint8 r,g,b,a;
		int x,y;

		FILE *file=fopen(filename,"wb");

		if(!file)
		{
			throw std::string("Could not open file \"")+std::string(filename)+std::string("\" for writing.");
		}

		png_ptr=png_create_write_struct(PNG_LIBPNG_VER_STRING,0,0,0);

		if(png_ptr==0)
		{
			fclose(file);
			png_destroy_write_struct(&png_ptr,(png_infopp)0);

			throw std::string("Could not create libpng data structures.");
		}
		else
		{
			info_ptr=png_create_info_struct(png_ptr);

			if(info_ptr==0)
			{
				fclose(file);
				png_destroy_write_struct(&png_ptr,(png_infopp)0);

				throw std::string("Could not create libpng data structures.");
			}
			else
			{
				if(setjmp(png_jmpbuf(png_ptr)))
				{
					fclose(file);
					png_destroy_write_struct(&png_ptr,(png_infopp)NULL);

					throw std::string("Could not set libpng error callback.");
				}
				else
				{
					png_init_io(png_ptr,file);

					info_ptr->width=surf->w;
					info_ptr->height=surf->h;
					info_ptr->bit_depth=8;
					info_ptr->color_type=PNG_COLOR_TYPE_RGBA;
					info_ptr->interlace_type=1;
					info_ptr->valid=0;

					png_write_info(png_ptr,info_ptr);

					// --- Save the picture... ---
					unsigned char **png_rows;
					png_rows=new unsigned char *[surf->h];

					for(y=0; y<surf->h; y++)
					{
						png_rows[y]=new unsigned char[4*surf->w];

						for(x=0; x<surf->w; x++)
						{
							SDL_GetRGBA(getPixel(surf,x,y),surf->format,&r,&g,&b,&a);

							png_rows[y][x*4+0]=r;
							png_rows[y][x*4+1]=g;
							png_rows[y][x*4+2]=b;
							png_rows[y][x*4+3]=a;
						}
					}

					png_write_image(png_ptr,png_rows);

					for(y=0; y<surf->h; y++)
					{
						free(png_rows[y]);
					}

					free(png_rows);
					png_write_end(png_ptr, NULL);
					png_destroy_write_struct(&png_ptr, &info_ptr);

					fclose(file);
				}
			}
		}
	}

	/**
	 * Creates a new background task. To run automatically, and for the task to be
	 * automatically deleted when it finishes, it must be added to the mysdl::backgroundTasks
	 * list.
	 *
	 * @param callback Callback object. Can be <code>0</code>.
	 */
	BackgroundTask::BackgroundTask(BackgroundTaskCallback *callback) :
		callback(callback),hasFinished(0),finishedCount(0),totalCount(0)
	{
		if(callback)
		{
			callback->onStart(this);
		}
	}

	/**
	 * Called when the task has finished.
	 */
	void BackgroundTask::finish()
	{
		hasFinished=true;

		if(callback)
		{
			callback->onFinish(this);
		}
	}

	/**
	 * Queries task progress.
	 *
	 * @param finished Will contain the number of finished steps for this task.
	 * @param total    Will contain the total number of steps for this task.
	 */
	void BackgroundTask::getProgress(Uint32 *finished,Uint32 *total) const
	{
		*finished=finishedCount;
		*total=totalCount;
	}

	/**
	 * Returns the relative task progress. The returned value is between 0 (no progress)
	 * and 1 (task finished).
	 *
	 * @return Relative progress.
	 */
	GLfloat BackgroundTask::getProgressRelative() const
	{
		return((GLfloat)(finishedCount)/(GLfloat)(totalCount));
	}

	/**
	 * Performs background task steps.
	 *
	 * @param timeLimit When this system time is reached, the background task will
	 *                  be suspended and mysdl::perform() will return. Background tasks
	 *                  will continue to run at the next call to mysdl::perform().
	 */
	void perform(Uint32 timeLimit)
	{
		static std::list<BackgroundTask *> finished;

		for(std::list<BackgroundTask *>::iterator itr=backgroundTasks.begin(); itr!=backgroundTasks.end(); itr++)
		{
			if(SDL_GetTicks()>=timeLimit)
			{
				break;
			}

			(*itr)->perform(timeLimit);

			if((*itr)->hasFinished)
			{
				finished.push_back(*itr);
			}
		}

		if(!finished.empty())
		{
			for(std::list<BackgroundTask *>::iterator itr=finished.begin(); itr!=finished.end(); itr++)
			{
				backgroundTasks.remove(*itr);
				delete *itr;
			}

			finished.clear();
		}
	}

	std::list<basic_Vector2D<myint> > getVideoModes()
	{
		std::list<basic_Vector2D<myint> > modes;

		SDL_Rect **sdlModes=SDL_ListModes(0,SDL_FULLSCREEN);

		if(sdlModes==(SDL_Rect **)-1)
		{
			debug::warning(dbgCore,"Unrestricted modes not yet handled correctly!");
		}
		else if(sdlModes)
		{
			for(int i=0; sdlModes[i]; i++)
			{
				modes.push_front(basic_Vector2D<myint>(sdlModes[i]->w,sdlModes[i]->h));
			}
		}

		return modes;
	}

	void setBlendMode(BlendMode mode)
	{
		switch(mode)
		{
			case bmDefault:
			case bmNormal:
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
				break;

			case bmAdd:
				glBlendFunc(GL_SRC_ALPHA,GL_ONE);
				break;

			default:
				assert(0);
		}
	}
};
