/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <iostream>

#include "controls.h"
#include "mysdl.h"

namespace mysdl
{
	static const std::string XML_RESOURCES_RESOURCES="resources";

	std::list<const ResourceBankGenerator *> *getGenerators()
	{
		static std::list<const ResourceBankGenerator *> generators;
		return(&generators);
	}

	void registerResourceBankGenerator(const ResourceBankGenerator *generator)
	{
		getGenerators()->push_back(generator);
	}

	void unregisterResourceBankGenerator(const ResourceBankGenerator *generator)
	{
		getGenerators()->remove(generator);
	}

	ResourceBankGenerator::ResourceBankGenerator()
	{
		registerResourceBankGenerator(this);
	}

	ResourceBankGenerator::~ResourceBankGenerator()
	{
//		unregisterResourceBankGenerator(this);
		/** \todo don't we ever need to delete resource bank generators? perhaps not */
	}

	class BasicResourceBankGenerator : public ResourceBankGenerator
	{
		public:
			BasicResourceBank *generateBank(const std::string &name) const;
	};

	BasicResourceBankGenerator resourceBankGenerator;

	BasicResourceBank *BasicResourceBankGenerator::generateBank(const std::string &name) const
	{
		if(name=="inputEvaluator")
		{
			return(new InputEvaluator());
		}
		else
		{
			return(0);
		}
	}

	BasicResourceBank* ResourceManager::createResourceBank(const std::string &name)
	{
		BasicResourceBank *bank=0;

		for(std::list<const ResourceBankGenerator *>::const_iterator itr=getGenerators()->begin(); itr!=getGenerators()->end(); itr++)
		{
			bank=(*itr)->generateBank(name);

			if(bank)
			{
				break;
			}
		}

		return(bank);
	}

	void ResourceManager::getFromNode(const xml::Node *root,const std::string &baseDir)
	{
		for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
		{
			xml::Node *node=*itr;

			std::string name=node->getAttribute("name","");
			std::string dir=baseDir+node->getAttribute("dir","");

			if(name.size()==0)
			{
				debug::error(dbgResources,"Missing name attribute from resource bank node.");
				continue;
			}

			BasicResourceBank *bank=createResourceBank(node->name);

			if(bank)
			{
				add(name,bank);
				bank->getFromNode(node,dir);
			}
			else
			{
				debug::warning(dbgResources,"Unknown resource bank type \"%s\".",node->name.c_str());
			}
		}
	}

	ResourceLoader::ResourceLoader(ResourceLoaderHandler *handler) :
		list(),handler(handler)
	{
	}

	ResourceLoader::~ResourceLoader()
	{
	}

	void ResourceLoader::operator+=(RefreshingBank *add)
	{
		list.push_back(add);
	}

	int loaderThreadStarter(void *loader)
	{
		((ResourceLoader *)(loader))->thread();
		return(0);
	}

	void ResourceLoader::start()
	{
		progress=total=0;

		for(std::list<RefreshingBank *>::iterator itr=list.begin(); itr!=list.end(); itr++)
		{
			total+=(*itr)->size();
		}

		handler->loaderUpdate(0,progress,total);
		thread();
	}

	void ResourceLoader::thread()
	{
		for(std::list<RefreshingBank *>::iterator itr=list.begin(); itr!=list.end(); itr++)
		{
			current=*itr;
			current->refresh(this);
		}

		handler->loaderFinished();
		delete(this);
	}

	void ResourceLoader::elementRefreshed()
	{
		progress++;
		handler->loaderUpdate(current,progress,total);
	}
};
